@@ -6,6 +6,8 @@ import fnmatch
 from datetime import datetime
 
 from .backends import S3Backend
+from .backends import RackspaceBackend
+from .backends import GoogleBackend
 
 import git
 import boto
@@ -43,14 +45,32 @@ def default_backend():
     except git.exc.GitCommandError:
         backend_name = None
 
-    if backend_name == "s3":
+    backend = backend_for_name(backend_name)
+
+    if backend:
+        return backend
+    else:
+        sys.stderr.write("error: s3, google, and rackspace are currently the only supported backends")
+        sys.exit(0)
+
+def backend_for_name(name):
+    if name == "s3":
         access_key_id = g.config("bigstore.s3.key", file=".bigstore")
         secret_access_key = g.config("bigstore.s3.secret", file=".bigstore")
         bucket_name = g.config("bigstore.s3.bucket", file=".bigstore")
         return S3Backend(access_key_id, secret_access_key, bucket_name)
+    elif name == "rackspace":
+        username = g.config("bigstore.rackspace.username", file=".bigstore")
+        api_key = g.config("bigstore.rackspace.key", file=".bigstore")
+        container_name = g.config("bigstore.rackspace.container", file=".bigstore")
+        return RackspaceBackend(username, api_key, container_name)
+    elif name == "google":
+        access_key_id = g.config("bigstore.google.key", file=".bigstore")
+        secret_access_key = g.config("bigstore.google.secret", file=".bigstore")
+        bucket_name = g.config("bigstore.google.bucket", file=".bigstore")
+        return GoogleBackend(access_key_id, secret_access_key, bucket_name)
     else:
-        sys.stderr.write("error: s3 is currently the only supported backend")
-        sys.exit(0)
+        return None
 
 def object_directory(hash_function_name):
     return os.path.join(git_directory, "bigstore", "objects", hash_function_name)
@@ -70,7 +90,10 @@ def mkdir_p(path):
 def upload_callback(filename):
     def inner(size, total):
         sys.stderr.write("\r")
-        sys.stderr.write("{: <4.0%}\t{}".format(size / float(total), filename))
+        if total > 0:
+            sys.stderr.write("{: <4.0%}\t{}".format(size / float(total), filename))
+        else:
+            sys.stderr.write("?%\t{}".format(filename))
 
     return inner
 
@@ -117,13 +140,13 @@ def push():
             # No notes exist for this object
             entries = []
 
+        backend = default_backend()
         for entry in entries:
-            if "upload" in entry:
+            if "upload" in entry and backend.name in entry:
                 break
         else:
             firstline, hash_function_name, hash = g.show(sha).split('\n')
             if firstline == 'bigstore':
-                backend = default_backend()
                 if not backend.exists(hash):
                     with open(object_filename(hash_function_name, hash)) as file:
                         backend.push(file, hash, cb=upload_callback(filename))
@@ -132,7 +155,7 @@ def push():
 
                     user_name = g.config("user.name")
                     user_email = g.config("user.email")
-                    g.notes("--ref=bigstore", "append", sha, "-m", "{}	upload	s3	{} <{}>".format(time.time(), user_name, user_email))
+                    g.notes("--ref=bigstore", "append", sha, "-m", "{}	upload	{}	{} <{}>".format(time.time(), backend.name, user_name, user_email))
 
     sys.stderr.write("pushing bigstore metadata...")
     g.push("origin", "refs/notes/bigstore")
@@ -156,14 +179,15 @@ def pull():
             entries = []
 
         for entry in entries:
-            if "upload" in entry.split('\t'):
+            timestamp, action, backend_name, _ = entry.split('\t')
+            if action == "upload":
                 firstline, hash_function_name, hash = g.show(sha).split('\n')
                 if firstline == 'bigstore':
                     try:
                         with open(object_filename(hash_function_name, hash)):
                             pass
                     except IOError:
-                        backend = default_backend()
+                        backend = backend_for_name(backend_name)
                         if backend.exists(hash):
                             with open(filename, 'wb') as file:
                                 backend.pull(file, hash, cb=upload_callback(filename))
@@ -172,7 +196,7 @@ def pull():
 
                             user_name = g.config("user.name")
                             user_email = g.config("user.email")
-                            g.notes("--ref=bigstore", "append", sha, "-m", "{}	download	s3	{} <{}>".format(time.time(), user_name, user_email))
+                            g.notes("--ref=bigstore", "append", sha, "-m", "{}	download	{}	{} <{}>".format(time.time() + time.timezone, backend.name, user_name, user_email))
                             g.add(filename)
 
                 break
@@ -182,19 +206,22 @@ def pull():
     sys.stderr.write("done\n")
 
 def filter_clean():
-    file = tempfile.NamedTemporaryFile(mode='w+t', delete=False)
-    hash = default_hash_function()
-
-    for line in sys.stdin:
-        if line == "bigstore\n":
+    firstline = sys.stdin.next()
+    if firstline == "bigstore\n":
+        sys.stdout.write(firstline)
+        for line in sys.stdin:
             sys.stdout.write(line)
-            sys.stdout.write(sys.stdin.next())
-            sys.stdout.write(sys.stdin.next())
-            break
-
-        hash.update(line)
-        file.write(line)
     else:
+        file = tempfile.NamedTemporaryFile(mode='w+t', delete=False)
+        hash = default_hash_function()
+
+        hash.update(firstline)
+        file.write(firstline)
+
+        for line in sys.stdin:
+            hash.update(line)
+            file.write(line)
+
         file.close()
 
         hexdigest = hash.hexdigest()
@@ -207,29 +234,45 @@ def filter_clean():
 
 
 def filter_smudge():
-    for line in sys.stdin:
-        if line == "bigstore\n":
-            hash_function_name = sys.stdin.next()
-            hash = sys.stdin.next()
-            source_filename = object_filename(hash_function_name[:-1], hash[:-1])
+    firstline = sys.stdin.next()
+    if firstline == "bigstore\n":
+        hash_function_name = sys.stdin.next()
+        hash = sys.stdin.next()
+        source_filename = object_filename(hash_function_name[:-1], hash[:-1])
 
-            try:
-                with open(source_filename):
-                    pass
-            except IOError:
-                sys.stdout.write(line)
-                sys.stdout.write(hash_function_name)
-                sys.stdout.write(hash)
-            else:
-                with open(source_filename, 'rb') as file:
-                    for line in file:
-                        sys.stdout.write(line)
+        try:
+            with open(source_filename):
+                pass
+        except IOError:
+            sys.stdout.write(firstline)
+            sys.stdout.write(hash_function_name)
+            sys.stdout.write(hash)
+        else:
+            with open(source_filename, 'rb') as file:
+                for line in file:
+                    sys.stdout.write(line)
+    else:
+        sys.stdout.write(firstline)
+        for line in sys.stdin:
+            sys.stdout.write(line)
 
-                break
+def request_rackspace_credentials():
+    print
+    print "Enter your Rackspace Cloud Files Credentials"
+    print
+    username = raw_input("Username: ")
+    api_key = raw_input("API Key: ")
+    container = raw_input("Container: ")
+
+    g.config("bigstore.backend", "rackspace", file=".bigstore")
+    g.config("bigstore.rackspace.username", username, file=".bigstore")
+    g.config("bigstore.rackspace.key", api_key, file=".bigstore")
+    g.config("bigstore.rackspace.container", container, file=".bigstore")
 
 def request_s3_credentials():
-    print "Please enter your S3 Credentials"
-    print ""
+    print
+    print "Enter your Amazon S3 Credentials"
+    print
     s3_key = raw_input("Access Key: ")
     s3_secret = raw_input("Secret Key: ")
     s3_bucket = raw_input("Bucket Name: ")
@@ -239,6 +282,19 @@ def request_s3_credentials():
     g.config("bigstore.s3.secret", s3_secret, file=".bigstore")
     g.config("bigstore.s3.bucket", s3_bucket, file=".bigstore")
 
+def request_google_cloud_storage_credentials():
+    print
+    print "Enter your Google Cloud Storage Credentials"
+    print
+    google_key = raw_input("Access Key: ")
+    google_secret = raw_input("Secret Key: ")
+    google_bucket = raw_input("Bucket Name: ")
+
+    g.config("bigstore.backend", "google", file=".bigstore")
+    g.config("bigstore.google.key", google_key, file=".bigstore")
+    g.config("bigstore.google.secret", google_secret, file=".bigstore")
+    g.config("bigstore.google.bucket", google_bucket, file=".bigstore")
+
 def log():
     filename = sys.argv[2]
     trees = g.log("--pretty=format:%T", filename).split('\n')
@@ -255,25 +311,46 @@ def log():
             timestamp, action, backend, user = note.split('\t')
             dt = datetime.fromtimestamp(float(timestamp))
             if action == "upload":
-                print "{}: uploaded to {} by {}".format(dt, backend, user)
+                print u"{}: {} \u2190 {}".format(dt.strftime("%a %b %e %X %Y +0000"), backend, user)
             else:
-                print "{}: downloaded from {} by {}".format(dt, backend, user)
+                print u"{}: {} \u2192 {}".format(dt.strftime("%a %b %e %X %Y +0000"), backend, user)
 
 def init():
     try:
-        with open(".bigstore"):
-            pass
-    except IOError:
-        request_s3_credentials()
+        g.config("bigstore.backend", file=".bigstore")
+    except git.exc.GitCommandError:
+        print "What backend would you like to store your files with?"
+        print "(1) Amazon S3"
+        print "(2) Google Cloud Storage"
+        print "(3) Rackspace Cloud Files"
+        choice = None
+        while choice not in ["1", "2", "3"]:
+            choice = raw_input("Enter your choice here: ")
+
+        if choice == "1":
+            try:
+                g.config("bigstore.s3.key", file=".bigstore")
+                g.config("bigstore.s3.secret", file=".bigstore")
+                g.config("bigstore.s3.bucket", file=".bigstore")
+            except git.exc.GitCommandError:
+                request_s3_credentials()
+        elif choice == "2":
+            try:
+                g.config("bigstore.google.key", file=".bigstore")
+                g.config("bigstore.google.secret", file=".bigstore")
+                g.config("bigstore.google.bucket", file=".bigstore")
+            except git.exc.GitCommandError:
+                request_google_cloud_storage_credentials()
+        elif choice == "3":
+            try:
+                g.config("bigstore.rackspace.username", file=".bigstore")
+                g.config("bigstore.rackspace.key", file=".bigstore")
+                g.config("bigstore.rackspace.container", file=".bigstore")
+            except git.exc.GitCommandError:
+                request_rackspace_credentials()
+
     else:
-        try:
-            g.config("bigstore.s3.key", file=".bigstore")
-            g.config("bigstore.s3.secret", file=".bigstore")
-            g.config("bigstore.s3.bucket", file=".bigstore")
-        except git.exc.GitCommandError:
-            request_s3_credentials()
-        else:
-            print "Reading credentials from .bigstore configuration file."
+        print "Reading credentials from .bigstore configuration file."
 
     try:
         g.fetch("origin", "refs/notes/bigstore:refs/notes/bigstore-remote", "--force")
